from telethon import types

from ..utils.filters import TelethonFilter

__all__ = [
    "is_message",
    "is_user_message",
    "is_channel_message",
    "is_by_user_message",
    "is_by_channel_message",
]

is_message = TelethonFilter(self__isinstance=types.Message)
is_user_message = TelethonFilter(peer_id__isinstance=types.PeerUser)
is_channel_message = TelethonFilter(peer_id__isinstance=types.PeerChannel)
is_by_user_message = TelethonFilter(from_id__isinstance=types.PeerUser)
is_by_channel_message = TelethonFilter(from_id__isinstance=types.PeerChannel)



import asyncio
import random
import typing as tp

__all__ = ["delay"]


@tp.overload
async def delay(seconds: float, *, disp: float = None) -> None:
    pass


@tp.overload
async def delay(min_value: float, max_value: float) -> None:
    pass


async def delay(x: float, y: float = None, *, disp: float = None) -> None:  # type: ignore[misc]
    if y is None:
        if disp is not None:
            y = x * (1 + disp)
        else:
            return await asyncio.sleep(x)

    await asyncio.sleep(random.uniform(x, y))



# flake8: noqa
# mypy: ignore-errors

from .timeout import *
from .rate_limit import *
from .serialization import *
from .filters import *

__all__ = timeout.__all__ + rate_limit.__all__ + serialization.__all__ + filters.__all__



import typing as tp

from ..errors import AgentAuthError

if tp.TYPE_CHECKING:
    from .agent import TelegramAgent


class AuthMixin:
    async def start_session(self: "TelegramAgent") -> None:
        if not self.client.is_connected():
            await self.client.connect()

        if not await self.client.is_user_authorized():
            raise AgentAuthError("User is not authorized")



import functools
import operator
import typing as tp

__all__ = ["TelethonFilter"]

Operator = tp.Callable[[tp.Any, tp.Any], bool]
Path = list[str]
Value = tp.Any


class TelethonFilter:
    _operator_postfix_func_mapping: dict[str, Operator] = {
        "eq": operator.eq,
        "ne": operator.ne,
        "is": operator.is_,
        "is_not": operator.is_not,
        "lt": operator.lt,
        "gt": operator.gt,
        "lte": operator.le,
        "gte": operator.ge,
        "isinstance": isinstance,
    }
    _default_operator_postfix = "eq"
    _path_separator = "__"
    _self_prefix = "self"

    def __init__(self, **criteria: Value) -> None:
        self._criteria = criteria
        self._compiled_criteria = self.compile(self._criteria)

    def __call__(self, obj: tp.Any) -> bool:
        for paths, op, value in self._compiled_criteria:
            curr_obj = obj
            for attr in paths:
                if attr != self._self_prefix:
                    curr_obj = getattr(curr_obj, attr, None)
                if curr_obj is None:
                    return False
            if not op(curr_obj, value):
                return False
        return True

    @classmethod
    def compile(cls, criteria: tp.Mapping[str, Value]) -> list[tuple[Path, Operator, Value]]:
        compiled_filters = []

        for key, value in criteria.items():
            path, op = key.split(cls._path_separator), cls._default_operator_postfix
            if path[-1] in cls._operator_postfix_func_mapping:
                *path, op = path
            compiled_filters.append((path, cls._operator_postfix_func_mapping[op], value))

        return compiled_filters

    def add(self, **criteria: Value) -> tp.Self:
        return self.__class__(**(self._criteria | criteria))

    def agg(self, *filters: "TelethonFilter") -> tp.Self:
        return self.__class__(**functools.reduce(lambda acc, item: acc | item.criteria, filters, self._criteria))

    @property
    def criteria(self) -> dict[str, Value]:
        return self._criteria



from . import constants, errors, filters, structures, utils
from .agent import TelegramAgent

__all__ = [
    "TelegramAgent",
    "constants",
    "errors",
    "filters",
    "structures",
    "utils",
]



import base64
import typing as tp
import zlib

from telethon.extensions import BinaryReader
from telethon.tl import TLObject

__all__ = ["TelethonString"]


class TelethonString:
    _encoding = "ascii"

    def __init__(self, string: str) -> None:
        self._string = string

    @classmethod
    def from_tl_object(cls, tl_object: TLObject) -> tp.Self:
        return cls(base64.b64encode(zlib.compress(bytes(tl_object))).decode(cls._encoding))

    def to_tl_object(self) -> TLObject:
        return BinaryReader(zlib.decompress(base64.b64decode(self._string))).tgread_object()

    def __str__(self) -> str:
        return self._string



import abc
import functools
import math
import time
import typing as tp
from contextlib import asynccontextmanager
from datetime import timedelta

from teleagent.utils import delay

try:
    from redis.asyncio import Redis
except ImportError:
    Redis = None  # type: ignore[misc,assignment]

__all__ = [
    "RateLimitMemory",
    "RateLimitRedis",
    "RateLimitHandler",
]

_F = tp.TypeVar("_F", bound=tp.Callable[..., tp.Awaitable[tp.Any]])


class RateLimit(abc.ABC):
    def __init__(self, function_name: str, ex_time: int | timedelta, **kwargs: tp.Any) -> None:
        self._function_name = function_name

        if isinstance(ex_time, timedelta):
            self._ex_time = int(ex_time.total_seconds())
        else:
            self._ex_time = ex_time

    @abc.abstractmethod
    async def set(self, slug: str) -> None:
        pass

    @abc.abstractmethod
    async def get_seconds(self, slug: str) -> int | None:
        pass


class RateLimitMemory(RateLimit):
    def __init__(self, *args: tp.Any, **kwargs: tp.Any) -> None:
        super().__init__(*args, **kwargs)

        self._storage: dict[str, float] = {}

    async def set(self, slug: str) -> None:
        expires_at = time.time() + self._ex_time
        self._storage[slug] = expires_at

    async def get_seconds(self, slug: str) -> int | None:
        current_time = time.time()
        expires_at = self._storage.get(slug)

        if expires_at is None:
            return None

        if current_time > expires_at:
            del self._storage[slug]
            return None

        return math.ceil(expires_at - current_time)


class RateLimitRedis(RateLimit):
    _key_prefix: str = "rate_limit"

    def __init__(self, *args: tp.Any, redis_client: Redis, **kwargs: tp.Any) -> None:
        super().__init__(*args, **kwargs)

        self._redis_client = redis_client

    async def set(self, slug: str) -> None:
        await self._redis_client.set(self._build_key(slug), 0, ex=self._ex_time)

    async def get_seconds(self, slug: str) -> int | None:
        ttl = await self._redis_client.ttl(self._build_key(slug))
        return ttl if ttl > 0 else None

    def _build_key(self, slug: str) -> str:
        return f"{self._key_prefix}:{self._function_name}:{slug}"


class RateLimitHandler:
    def __init__(self, slug: str) -> None:
        self._slug = slug

    @asynccontextmanager
    async def handle(self, rate_limit: RateLimit) -> tp.AsyncGenerator[None, None]:
        seconds = await rate_limit.get_seconds(self._slug)
        if seconds is not None:
            await delay(seconds)

        try:
            yield
        finally:
            await rate_limit.set(self._slug)

    def decorate(self, rate_limit: RateLimit) -> tp.Callable[[_F], _F]:
        def decorator(func: _F) -> _F:
            @functools.wraps(func)
            async def wrapper(*args: tp.Any, **kwargs: tp.Any) -> tp.Any:
                async with self.handle(rate_limit):
                    return await func(*args, **kwargs)

            return tp.cast(_F, wrapper)

        return decorator



from telethon import types

from ..utils.filters import TelethonFilter

__all__ = [
    "is_channel",
    "is_group_channel",
]

is_channel = TelethonFilter(self__isinstance=types.Channel)
is_group_channel = TelethonFilter(megagroup__is=True)



import abc
import typing as tp
from dataclasses import asdict

from telethon import TelegramClient, sessions

from ..structures import AppData, ProxyData

if tp.TYPE_CHECKING:
    from .agent import TelegramAgent


class TelegramAgentBase(abc.ABC):
    def __init__(
        self: "TelegramAgent",
        session: sessions.Session,
        api_id: int,
        api_hash: str,
        *,
        app: AppData | None = None,
        proxy: ProxyData | None = None,
        **kwargs: tp.Any,
    ) -> None:
        telegram_client_kwargs: dict[str, tp.Any] = {}

        if app is not None:
            kwargs.update(asdict(app))
            kwargs.pop("lang_pack")

        if proxy is not None:
            kwargs["proxy"] = asdict(proxy)

        telegram_client_kwargs.update(kwargs)

        self._client = TelegramClient(session, api_id, api_hash, **telegram_client_kwargs)

        if app is not None:
            self._client._init_request.lang_pack = app.lang_pack  # noqa

    @property
    def client(self: "TelegramAgent") -> TelegramClient:
        return self._client
